module env = import "../futhark_mpi/envelope"

type envelope_1d_t [tx] = env.envelope_1d_t [tx]
type envelope_2d_t [n][m][ty][tx] = env.envelope_2d_t [n][m][ty][tx]
type~ side_envelope_3d_t = env.side_envelope_3d_t
type~ envelope_3d_t = env.envelope_3d_t

type~ envelope_1d_i8 = {
    east: []i8,    -- 0
    west: []i8     -- 1
}

entry get_envelope_1d [n] (xs: [n]u8) (thickness: i64) : envelope_1d_t [thickness] =
    env.get_envelope_1d xs thickness

entry get_envelope_2d [n][m] (matrix: [n][m]u8) (thickness_y: i64) (thickness_x: i64): envelope_2d_t [n][m][thickness_y][thickness_x] =
    env.get_envelope_2d matrix thickness_y thickness_x

entry get_envelope_3d [n][m][l] (cube: [n][m][l]u8) (thickness_y: i64) (thickness_x: i64) (thickness_z: i64): envelope_3d_t =
    env.get_envelope_3d cube thickness_y thickness_x thickness_z

entry augment_1d [n][tx] (xs: [n]u8) (envelope: envelope_1d_t [tx]) :[]u8 = env.augment_1d xs envelope
entry augment_2d [n][m][ty][tx] (matrix: [n][m]u8) (envelope: envelope_2d_t [n][m][tx][ty]) :[][]u8 = env.augment_2d matrix envelope
entry augment_3d [n][m][l] (cube: [n][m][l]u8) (envelope: envelope_3d_t) :[][][]u8 = env.augment_3d cube envelope

let get_neighbours [n] (elem: [n]i8) : [n](i8,i8) =
    let east = rotate 1 elem
    let west = rotate (-1) elem
    in zip2 west east

entry next_chunk_elems [n] (chunk_elems :[n]i8) :[]i8 =
    let neighbours = get_neighbours chunk_elems
    let next_elems = map2 (\cell neighbours ->
                        let west = neighbours.0
                        let east = neighbours.1
                        let symbol = (west << 2) | (cell << 1) | east
                        in i8.bool (symbol >= 1 && symbol <= 4)
                     ) chunk_elems neighbours
    in next_elems[1:n-1]
