type envelope_1d_t [tx] = {
    east: [tx]u8,  -- 0
    west: [tx]u8   -- 1
}

type envelope_2d_t [n][m][ty][tx] = {
    east: [n][tx]u8,         -- 0
    north: [ty][m]u8,        -- 1
    north_east: [ty][tx]u8,  -- 2
    north_west: [ty][tx]u8,  -- 3
    south: [ty][m]u8,        -- 4
    south_east: [ty][tx]u8,  -- 5
    south_west: [ty][tx]u8,  -- 6
    west: [n][tx]u8          -- 7
}

type~ side_envelope_3d_t = {
    east: [][][]u8,       -- edge
    north: [][][]u8,      -- edge
    north_east: [][][]u8, -- vertice
    north_west: [][][]u8, -- vertice
    south: [][][]u8,      -- edge
    south_east: [][][]u8, -- vertice
    south_west: [][][]u8, -- vertice
    surface: [][][]u8,
    west: [][][]u8        -- edge
}

type~ envelope_3d_t = {
    back: side_envelope_3d_t,   -- 0
    bottom: side_envelope_3d_t, -- 1
    front: side_envelope_3d_t,  -- 2
    left: side_envelope_3d_t,   -- 3
    right: side_envelope_3d_t,  -- 4
    top: side_envelope_3d_t     -- 5
}

entry get_envelope_1d [n] (xs: [n]u8) (thickness: i64) : envelope_1d_t [thickness] =
    let west = xs[0:thickness] :> [thickness]u8
    let east = xs[n - thickness:] :> [thickness]u8
    in {east, west}

entry get_envelope_2d [n][m] (matrix: [n][m]u8) (thickness_y: i64) (thickness_x: i64): envelope_2d_t [n][m][thickness_y][thickness_x] =
    let north = matrix[0:thickness_y] :> [thickness_y][m]u8

    let north_west = matrix[0:thickness_y,0:thickness_x] :> [thickness_y][thickness_x]u8
    let north_east = matrix[0:thickness_y,m - thickness_x:m] :> [thickness_y][thickness_x]u8

    let south = matrix[n - thickness_y:] :> [thickness_y][m]u8

    let south_east = matrix[n - thickness_y:n,m - thickness_x:m] :> [thickness_y][thickness_x]u8
    let south_west = matrix[n - thickness_y:n,0:thickness_x] :> [thickness_y][thickness_x]u8

    let east = matrix[:,m - (thickness_x):] :> [n][thickness_x]u8
    let west = matrix[:,0:thickness_x] :> [n][thickness_x]u8

    in {east, north, north_east, north_west, south, south_east, south_west, west}

entry get_envelope_3d [n][m][l] (cube:[n][m][l]u8) (thickness_y: i64) (thickness_x: i64) (thickness_z: i64): envelope_3d_t =
    let front = {
        north_west = cube[0:thickness_y, 0:thickness_x, 0:thickness_z] :> [thickness_y][thickness_x][thickness_z]u8,
        west = cube[:, 0:thickness_x, 0:thickness_z] :> [n][thickness_x][thickness_z]u8,
        south_west = cube[n - thickness_y:, 0:thickness_x, 0:thickness_z] :> [thickness_y][thickness_x][thickness_z]u8,
        south = cube[n - thickness_y: ,:, 0:thickness_z] :>  [thickness_y][m][thickness_z]u8,
        south_east = cube[n - thickness_y:n, m - thickness_x:, 0:thickness_z] :> [thickness_y][thickness_x][thickness_z]u8,
        east = cube[:, m - thickness_x:, 0:thickness_z] :> [n][thickness_x][thickness_z]u8,
        north_east = cube[0:thickness_y, m - thickness_x:, 0:thickness_z] :> [thickness_y][thickness_x][thickness_z]u8,
        north = cube[0:thickness_y   ,:, 0:thickness_z] :>  [thickness_y][m][thickness_z]u8,
        surface = cube[:, :, 0:thickness_z] :> [n][m][thickness_z]u8
    }
    let back = {
        north_west = cube[0:thickness_y, 0:thickness_x, l-thickness_z:] :> [thickness_y][thickness_x][thickness_z]u8,
        west = cube[:, 0:thickness_x, l-thickness_z:] :> [n][thickness_x][thickness_z]u8,
        south_west = cube[n - thickness_y:, 0:thickness_x, l-thickness_z:] :> [thickness_y][thickness_x][thickness_z]u8,
        south = cube[n - thickness_y: ,:, l-thickness_z:] :>  [thickness_y][m][thickness_z]u8,
        south_east = cube[n - thickness_y:n, m - thickness_x:, l-thickness_z:] :> [thickness_y][thickness_x][thickness_z]u8,
        east = cube[:, m - thickness_x:, l-thickness_z:] :> [n][thickness_x][thickness_z]u8,
        north_east = cube[0:thickness_y, m - thickness_x:, l-thickness_z:] :> [thickness_y][thickness_x][thickness_z]u8,
        north = cube[0:thickness_y   ,:, l-thickness_z:] :>  [thickness_y][m][thickness_z]u8,
        surface = cube[:, :, l-thickness_z:] :> [n][m][thickness_z]u8
    }

    let top = {
        north_west = cube[0:thickness_y, 0:thickness_x, l-thickness_z:] :> [thickness_y][thickness_x][thickness_z]u8,
        west = cube[0:thickness_y, 0:thickness_x, :] :> [thickness_y][thickness_x][l]u8,
        south_west = cube[0:thickness_y, 0:thickness_x, 0:thickness_z] :> [thickness_y][thickness_x][thickness_z]u8,
        south = cube[0:thickness_y, :, 0:thickness_z] :> [thickness_y][m][thickness_z]u8,
        south_east = cube[0:thickness_y, m - thickness_x:, 0:thickness_z] :> [thickness_y][thickness_x][thickness_z]u8,
        east = cube[0:thickness_y, m-thickness_x:, :] :> [thickness_y][thickness_x][l]u8,
        north_east = cube[0:thickness_y, m - thickness_x:, l-thickness_z:] :> [thickness_y][thickness_x][thickness_z]u8,
        north = cube[0:thickness_y, :, l-thickness_z:] :> [thickness_y][m][thickness_z]u8,
        surface = cube[0:thickness_y, :, :] :> [thickness_y][m][l]u8
    }
    let bottom = {
        north_west = cube[n - thickness_y:, 0:thickness_x, l-thickness_z:] :> [thickness_y][thickness_x][thickness_z]u8,
        west = cube[n-thickness_y:, 0:thickness_x, :] :> [thickness_y][thickness_x][l]u8,
        south_west = cube[n - thickness_y:, 0:thickness_x, 0:thickness_z] :> [thickness_y][thickness_x][thickness_z]u8,
        south = cube[n-thickness_y:, :, 0:thickness_z] :> [thickness_y][m][thickness_z]u8,
        south_east = cube[n - thickness_y:n, m - thickness_x:, 0:thickness_z] :> [thickness_y][thickness_x][thickness_z]u8,
        east = cube[n-thickness_y:, m-thickness_x:, :] :> [thickness_y][thickness_x][l]u8,
        north_east = cube[n - thickness_y:, m - thickness_x:, l-thickness_z:] :> [thickness_y][thickness_x][thickness_z]u8,
        north = cube[n-thickness_y:, :, l-thickness_z:] :> [thickness_y][m][thickness_z]u8,
        surface = cube[n-thickness_y:, :, :] :> [thickness_y][m][l]u8
    }

    let left = {
        north_west = cube[0:thickness_y, 0:thickness_x, l-thickness_z:] :> [thickness_y][thickness_x][thickness_z]u8,
        west = cube[:, 0:thickness_x, l-thickness_z:] :> [n][thickness_x][thickness_z]u8,
        south_west = cube[n - thickness_y:, 0:thickness_x, l-thickness_z:] :> [thickness_y][thickness_x][thickness_z]u8,
        south = cube[:, 0:thickness_x, l-thickness_z:] :> [n][thickness_x][thickness_z]u8,
        south_east = cube[n - thickness_y:, 0:thickness_x, 0:thickness_z] :> [thickness_y][thickness_x][thickness_z]u8,
        east = cube[:, 0:thickness_x, 0:thickness_z] :> [n][thickness_x][thickness_z]u8,
        north_east = cube[0:thickness_y, 0:thickness_x, 0:thickness_z] :> [thickness_y][thickness_x][thickness_z]u8,
        north =  cube[0:thickness_y, 0:thickness_x, :] :> [thickness_y][thickness_x][l]u8,
        surface = cube[:,0:thickness_x, :] :> [n][thickness_x][l]u8
    }

    let right = {
        north_west = cube[0:thickness_y, m - thickness_x:, 0:thickness_z] :> [thickness_y][thickness_x][thickness_z]u8,
        west = cube[:, m - thickness_x:, 0:thickness_z] :> [n][thickness_x][thickness_z]u8,
        south_west = cube[n - thickness_y:n, m - thickness_x:, 0:thickness_z] :> [thickness_y][thickness_x][thickness_z]u8,
        south = cube[n-thickness_y:, m-thickness_x:, :] :> [thickness_y][thickness_x][l]u8,
        south_east = cube[n - thickness_y:n, m - thickness_x:, l-thickness_z:] :> [thickness_y][thickness_x][thickness_z]u8,
        east = cube[:, m - thickness_x:, l-thickness_z:] :> [n][thickness_x][thickness_z]u8,
        north_east = cube[0:thickness_y, m - thickness_x:, l-thickness_z:] :> [thickness_y][thickness_x][thickness_z]u8,
        north = cube[0:thickness_y, m-thickness_x:, :] :> [thickness_y][thickness_x][l]u8,
        surface = cube[:,m-thickness_x:, :] :> [n][thickness_x][l]u8
    }

    in { back, bottom, front, left, right, top }

entry augment_1d [n][tx] (xs: [n]u8) (envelope: envelope_1d_t [tx]) : []u8 =
    let new_n = n + tx * 2
    in tabulate (new_n) (\i ->
        -- West
        if (i >= 0 && i < tx) then envelope.west[i]
        -- East
        else if (i >= (n + tx) && i < new_n) then envelope.east[i - n - tx]
        else xs[i - tx]
    ) :> [new_n]u8

entry augment_2d [n][m][ty][tx] (matrix:[n][m]u8) (envelope: envelope_2d_t [n][m][ty][tx]) :[][]u8 =
    let new_n = n + ty * 2
    let new_m = m + tx * 2
    in tabulate_2d (new_n) (new_m) (\i j ->
        -- North-West
        if (i >= 0 && i < ty && j >= 0 && j < tx) then envelope.north_west[i,j]
        -- North-East
        else if (i >= 0 && i < ty && j >= (m + tx) && j < new_m) then envelope.north_east[i,j - m - tx]
        -- South-West
        else if ( i >= (n + ty) && i < new_n && j >= 0 && j < tx ) then envelope.south_west[i-n-ty,j]
        -- South-East
        else if (i >= (n + ty) && i < new_n && j >= (m + tx) && j < new_m) then envelope.south_east[i - n - ty, j - m - tx]
        -- North
        else if ( i >= 0 && i < ty && j >= tx && j < (m + tx) ) then envelope.north[i, j - tx]
        -- West
        else if ( i >= ty && i < (n+ty) &&  j >= 0 && j < tx ) then envelope.west[i - ty, j]
        -- East
        else if ( i >= ty && i < (n+ty) &&  j >= m+tx && j < new_m ) then envelope.east[i - ty, j - m - tx]
        -- South
        else if ( i >= (n + ty) && i < new_n && j >= tx && j < (m + tx) ) then envelope.south[i - n - tx, j - tx]
        else matrix[i - ty, j - tx]
    ) :> [new_n][new_m]u8

entry augment_3d [n][m][l] (cube :[n][m][l]u8) (envelope: envelope_3d_t) :[][][]u8 =
    let ty = length envelope.top.surface
    let tx = length envelope.left.surface[0]
    let tz = length envelope.front.surface[0,0]
    let new_n = n + ty * 2
    let new_m = m + tx * 2
    let new_l = l + tz * 2
    in tabulate_3d (new_n) (new_m) (new_l) (\i j k ->
        -- Top-South-West
        if (i >= 0 && i < ty && j >= 0 && j < tx && k >= 0 && k < tz) then envelope.top.south_west[i,j,k]
        -- Top-West
        else if (i >= 0 && i < ty && j >= 0 && j < tx && k >= tz && k < l + tz) then envelope.top.west[i,j,k-tz]
        -- Top-North-West
        else if (i >= 0 && i < ty && j >= 0 && j < tx && k >= (l + tz) && k < new_l) then envelope.top.north_west[i,j,k-l-tz]
        -- Top-South
        else if (i >= 0 && i < ty && j >= tx && j < (m+tx) && k >= 0 && k < tz) then envelope.top.south[i,j-tx,k]
        -- Top-Surface
        else if (i >= 0 && i < ty && j >= tx && j < (m+tx) && k >= tz && k < (l+tz)) then envelope.top.surface[i,j-tx,k-tz]
        -- Top North
        else if (i >= 0 && i < ty && j >= tx && j < (m+tx) && k >= (l+tz) && k < new_l) then envelope.top.north[i,j-tx, k-l-tz]
        -- Top South East
        else if (i >= 0 && i < ty && j >= m+tx && j < new_m && k >= 0 && k < tz) then envelope.top.south_east[i, j-m-tx,k]
        -- Top East
        else if (i >= 0 && i < ty && j >= m+tx && j < new_m && k >= tz && k < l + tz) then envelope.top.east[i,j-m-tx,k-tz]
        -- Top North East
        else if (i >= 0 && i < ty && j >= m+tx && j < new_m && k >= l+tz && k < new_l) then envelope.top.north_east[i,j-m-tx,k-l-tz]

        -- Front-West
        else if (i >= ty && i < n+ty && j >= 0 && j < tx && k >= 0 && k < tz) then envelope.front.west[i-ty, j, k]
        -- Left
        else if (i >= ty && i < n+ty && j >= 0 && j < tx && k >= tz && k < l+tz) then envelope.left.surface[i-ty, j, k-tz]
        -- Back West
        else if (i >= ty && i < n+ty && j >= 0 && j < tx && k >= l+tz && k < new_l) then envelope.back.west[i-ty, j, k-l-tz]
        -- Front Surface
        else if (i >= ty && i < n+ty && j >= tx && j < m+tx && k >= 0 && k < tz) then envelope.front.surface[i-ty, j-tx, k]
        -- Back Surface
        else if (i >= ty && i < n+ty && j >= tx && j < m+tx && k >= l+tz && k < new_l) then envelope.back.surface[i-ty, j-tx, k-l-tz]
        -- Front East
        else if (i >= ty && i < n+ty && j >= m+tx && j < new_m && k >= 0 && k < tz) then envelope.front.east[i-ty, j-m-tx, k]
        -- Right
        else if (i >= ty && i < n+ty && j >= m+tx && j < new_m && k >= tz && k < l+tz) then envelope.right.surface[i-ty, j-m-tx, k-tz]
        -- Back East
        else if (i >= ty && i < n+ty && j >= m+tx && j < new_m && k >= l+tz && k < new_l) then envelope.back.east[i-ty, j-m-tx, k-l-tz]

        -- Bottom-South-West
        else if (i >= n+ty && i < new_n && j >= 0 && j < tx && k >= 0 && k < tz) then envelope.bottom.south_west[i-n-ty,j,k]
        -- Bottom-West
        else if (i >= n+ty && i < new_n && j >= 0 && j < tx && k >= tz && k < l + tz) then envelope.bottom.west[i-n-ty,j,k-tz]
        -- Bottom-North-West
        else if (i >= n+ty && i < new_n && j >= 0 && j < tx && k >= (l + tz) && k < new_l) then envelope.bottom.north_west[i-n-ty,j,k-l-tz]
        -- Bottom-South
        else if (i >= n+ty && i < new_n && j >= tx && j < (m+tx) && k >= 0 && k < tz) then envelope.bottom.south[i-n-ty,j-tx,k]
        -- Bottom-Surface
        else if (i >= n+ty && i < new_n && j >= tx && j < (m+tx) && k >= tz && k < (l+tz)) then envelope.bottom.surface[i-n-ty,j-tx, k-tz]
        -- Bottom North
        else if (i >= n+ty && i < new_n && j >= tx && j < (m+tx) && k >= (l+tz) && k < new_l) then envelope.bottom.north[i-n-ty,j-tx, k-l-tz]
        -- Bottom South East
        else if (i >= n+ty && i < new_n && j >= m+tx && j < new_m && k >= 0 && k < tz) then envelope.bottom.south_east[i-n-ty, j-m-tx,k]
        -- Bottom East
        else if (i >= n+ty && i < new_n && j >= m+tx && j < new_m && k >= tz && k < l + tz) then envelope.bottom.east[i-n-ty,j-m-tx,k-tz]
        -- Bottom North East
        else if (i >= n+ty && i < new_n && j >= m+tx && j < new_m && k >= l+tz && k < new_l) then envelope.bottom.north_east[i-n-ty,j-m-tx,k-l-tz]

        else cube[i-ty, j-tx, k-tz]
    ) :> [new_n][new_m][new_l]u8
